#include "internal.hpp"

namespace CaDiCaL {

// This function determines the next decision variable on the queue, without
// actually removing it from the decision queue, e.g., calling it multiple
// times without any assignment will return the same result.  This is of
// course used below in 'decide' but also in 'reuse_trail' to determine the
// largest decision level to backtrack to during 'restart' without changing
// the assigned variables (if 'opts.restartreusetrail' is non-zero).

int Internal::next_decision_variable_on_queue () {
  long searched = 0;
  int res = queue.unassigned;
  while (val (res))
    res = link (res).prev, searched++;
  if (searched) {
    stats.searched += searched;
    update_queue_unassigned (res);
  }
  LOG ("next queue decision variable %d bumped %ld", res, bumped (res));
  return res;
}

// This function determines the best decision with respect to score.
//
int Internal::next_decision_variable_with_best_score () {
  int res = 0;
  for (;;) {
    res = scores.front ();
    if (!val (res)) break;
    (void) scores.pop_front ();
  }
  LOG ("next decision variable %d with score %g", res, score (res));
  return res;
}

int Internal::next_decision_variable () {
  if (use_scores ()) return next_decision_variable_with_best_score ();
  else               return next_decision_variable_on_queue ();
}

// Implements phase saving as well using a target phase.

inline int Internal::decide_phase (int idx) {
  int phase = stable ? phases.target[idx] : 0;
  if (!phase) phase = pign (phases.saved[idx]);
  assert (phase);
  return phase * idx;
}

bool Internal::satisfied () {
  size_t assigned = trail.size ();
  assert (propagated == assigned);
  if ((size_t) level < assumptions.size ()) return false;
  return (assigned == (size_t) max_var);
}

// Search for the next decision and assign it to the saved phase.  Requires
// that not all variables are assigned.

int Internal::decide () {
  assert (!satisfied ());
  START (decide);
  int res = 0;
  if ((size_t) level < assumptions.size ()) {
    const int lit = assumptions[level];
    assert (assumed (lit));
    const int tmp = val (lit);
    if (tmp < 0) {
      LOG ("assumption %d falsified", lit);
      failing ();
      res = 20;
    } else if (tmp > 0) {
      LOG ("assumption %d already satisfied", lit);
      level++;
      control.push_back (Level (0, trail.size ()));
      LOG ("added pseudo decision level");
    } else {
      LOG ("deciding assumption %d", lit);
      search_assume_decision (lit);
    }
  } else {
    stats.decisions++;
    int idx = next_decision_variable ();
    int decision = decide_phase (idx);
    search_assume_decision (decision);
  }
  STOP (decide);
  return res;
}

}
